/*
PPKMouseAddon
 by Mnjul/purincess (Min-Zhong Lu)
 # Middle Sea, Mnjul's Intimate Home

Copyright 2007-2010 Min-Zhong Lu

This file is part of PPKMouseAddon.

PPKMouseAddon is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

PPKMouseAddon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PPKMouseAddon.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "stdafx.h"
#include "CallbackImpl.h"
#include "Settings.h"
#include "AboutDlg.h"
#include "OptionsDlg.h"

#include "resource.h"

LRESULT CallbackImpl::callWndProcInternal(int code, WPARAM wparam, LPARAM lparam){
	LPCWPSTRUCT pCWP = reinterpret_cast<LPCWPSTRUCT>(lparam);
	if(pCWP->message == WM_CREATE){
		wchar_t className[512] = {0};
		GetClassNameW(pCWP->hwnd, className, 511);
		// PCMan Main
		if(wcscmp(className, L"PCMan") == 0 || wcscmp(className, L"PCManCB") == 0 || wcscmp(className, L"PCManC") == 0){
			Settings::getInstance().loadSettings();

			hwndMain = pCWP->hwnd;
			previousMainWndProc = GetWindowLongPtr(pCWP->hwnd, GWLP_WNDPROC);
			SetWindowLongPtr(pCWP->hwnd, GWLP_WNDPROC, reinterpret_cast<LONG_PTR>(newMainWndProc));

		// PCMan BBSView
		}else if(wcscmp(className, L"BBS_View") == 0){
			previousTerminalViewWndProc = GetWindowLongPtr(pCWP->hwnd, GWLP_WNDPROC);
			SetWindowLongPtr(pCWP->hwnd, GWLP_WNDPROC, reinterpret_cast<LONG_PTR>(newTerminalViewWndProc));
		}
	}else if(pCWP->message == WM_DESTROY){
		wchar_t className[512] = {0};
		GetClassNameW(pCWP->hwnd, className, 511);
		if(wcscmp(className, L"PCMan") == 0 || wcscmp(className, L"PCManCB") == 0){
			Settings::getInstance().saveSettings();
		}
	}
	return CallNextHookEx(/* ignored */ NULL, code, wparam, lparam);
}

LRESULT CallbackImpl::newMainWndProcInternal(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam){
	LRESULT ret = 0;

	// for startup:
	// if you look at pcman's source code, it's obvious that the correct moment we can intercept and append a menu is when UpdateWindow,
	// and that is WM_PAINT. but intercepting WM_PAINT is too costly, so instead we'll use WM_SHOWWINDOW.
	// but showwindow only does not work for some people, so we add size orz.

	if(msg == WM_SHOWWINDOW || msg == WM_SIZE){
		ret = CallWindowProc(reinterpret_cast<WNDPROC>(previousMainWndProc), hwnd, msg, wparam, lparam);
		if(calledFirstTimeShowWindow.find(hwnd) == calledFirstTimeShowWindow.end()){
			if(reappendPCManMenu(hwnd)){
				calledFirstTimeShowWindow.insert(map_hwnd_bool::value_type(hwnd, true));
			}
		}
	// and we need to intercept the ID_CUSTOMIZE message...
	}else if(msg == WM_COMMAND){
		if(LOWORD(wparam) == PCMan_ID_CUSTOMIZE){
			ret = CallWindowProc(reinterpret_cast<WNDPROC>(previousMainWndProc), hwnd, msg, wparam, lparam);
			reappendPCManMenu(hwnd);
		}else if(LOWORD(wparam) == ID_PPK_ABOUT){
			AboutDlg dlgAbout(IDD_ABOUT, hwndMain);
			dlgAbout.DoModal();
		}else if(LOWORD(wparam) == ID_PPK_ENABLED){
			// get menu info, check checked, reverse it.
			MENUITEMINFO mii = {0};
			mii.cbSize = sizeof(MENUITEMINFO);
			mii.fMask = MIIM_STATE;
			GetMenuItemInfo(menuAppend, ID_PPK_ENABLED, FALSE, &mii);
			mii.fState ^= MFS_CHECKED;
			SetMenuItemInfo(menuAppend, ID_PPK_ENABLED, FALSE, &mii);

			if(mii.fState & MFS_CHECKED){
				Settings::getInstance().setEnabled(true);
			}else{
				Settings::getInstance().setEnabled(false);
			}
		}else if(LOWORD(wparam) == ID_PPK_OPTIONS){
			OptionsDlg dlgOptions(IDD_OPTIONS, hwndMain);
			dlgOptions.DoModal();
		}else{
			ret = CallWindowProc(reinterpret_cast<WNDPROC>(previousMainWndProc), hwnd, msg, wparam, lparam);
		}
	}else{
		ret = CallWindowProc(reinterpret_cast<WNDPROC>(previousMainWndProc), hwnd, msg, wparam, lparam);
	}
	return ret;
}

LRESULT CallbackImpl::newTerminalViewWndProcInternal(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam){
	if(!Settings::getInstance().getEnabled()){
		return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
	}

	// if it's ansi-editor mode, do nothing.
	HWND hwndMainTab = GetDlgItem(hwndMain, PCMan_IDC_MAINTAB);
	if(hwndMainTab){
		int cursel = TabCtrl_GetCurSel(hwndMainTab);
		TCITEM tci;
		tci.mask = TCIF_TEXT;
		wchar_t itemText[512] = {0};
		tci.pszText = &itemText[0];
		tci.cchTextMax = 511;
		TabCtrl_GetItem(hwndMainTab, cursel, &tci);
		if(wcsstr(itemText, L"ANSIms") != NULL){
			return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
		}
	}

	// for lbutton: if current cursor is hand, propagate it directly
	// but for some reason, if we GetCursor here, it's never IDC_HAND, so we need to record cursor info for mousemove
	if(msg == WM_MOUSEMOVE){
		LRESULT ret = CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
		POINT cursorPos;
		GetCursorPos(&cursorPos);
		HWND wndUnderCursor = WindowFromPoint(cursorPos);
		DWORD threadUnderCursor = GetWindowThreadProcessId(wndUnderCursor, NULL);
		DWORD threadCurrent = GetCurrentThreadId();
		if(threadCurrent != threadUnderCursor){
			AttachThreadInput(threadCurrent, threadUnderCursor, TRUE);
			currentCursor = GetCursor();
			AttachThreadInput(threadCurrent, threadUnderCursor, FALSE);
		}else{
			currentCursor = GetCursor();
		}
			
		return ret;
	// Lbutton -> right
	// for selection we record current position when LbuttonDown only
	// and when LbuttonUp, if mouse has gone > 10px, we don't send right and flip hasSelectionCurrently on
	}else if(msg == WM_LBUTTONDOWN){
		if(!Settings::getInstance().getEnabledMouseLeft() || currentCursor == LoadCursor(NULL, IDC_HAND) || currentCursor == LoadCursor(LoadLibrary(L"PCman.exe"), MAKEINTRESOURCE(PCMan_IDC_HAND_CUR))){
			return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
		}else{
			if(wparam & MK_CONTROL){
				return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam ^ MK_CONTROL, lparam);
			}else{
				mousePosition.x = GET_X_LPARAM(lparam);
				mousePosition.y = GET_Y_LPARAM(lparam);
				if(wparam & MK_RBUTTON){
					rbtnAndSomeButton = true;
					return 0;
				}else if(wparam & MK_MBUTTON){
					wbtnAndSomeButton = true;
					return 0;
				}else{
					lbtnAndSomeButton = false;
					return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam ^ MK_CONTROL, lparam);
				}
			}
		}
	}else if(msg == WM_LBUTTONUP){
		if(!Settings::getInstance().getEnabledMouseLeft() || currentCursor == LoadCursor(NULL, IDC_HAND) || currentCursor == LoadCursor(LoadLibrary(L"PCman.exe"), MAKEINTRESOURCE(PCMan_IDC_HAND_CUR))){
			return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
		}else{
			if(wparam & MK_CONTROL){
				return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam ^ MK_CONTROL, lparam);
			}else{
				int currentX = GET_X_LPARAM(lparam);
				int currentY = GET_Y_LPARAM(lparam);
				if((currentX - mousePosition.x) * (currentX - mousePosition.x) + (currentY - mousePosition.y) * (currentY - mousePosition.y) > 100){	// it's selection
					currentSelecting = true;
					return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
				}else{
					if(currentSelecting){
						// cancel selection
						currentSelecting = false;
						return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
					}else{
						if(wparam & MK_RBUTTON){
							keystroke(Settings::getInstance().getHKRL().vkeyCode, Settings::getInstance().getHKRL().modifiers);
							return 0;
						}
						if(wparam & MK_MBUTTON){
							keystroke(Settings::getInstance().getHKWL().vkeyCode, Settings::getInstance().getHKWL().modifiers);
							return 0;
						}

						if(!lbtnAndSomeButton){
							keybd_event((Settings::getInstance().getLRRL() == Settings::LR_TO_RL) ? VK_RIGHT : VK_LEFT, 0, 0, 0);
							keybd_event((Settings::getInstance().getLRRL() == Settings::LR_TO_RL) ? VK_RIGHT : VK_LEFT, 0, KEYEVENTF_KEYUP, 0);
						}

						return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
					}
				}
				return 0;
			}
		}
	// Rbutton -> Left
	}else if(msg == WM_RBUTTONDOWN){
		if(!Settings::getInstance().getEnabledMouseRight() || currentCursor==LoadCursor(NULL, IDC_HAND) || currentCursor == LoadCursor(LoadLibrary(L"PCman.exe"), MAKEINTRESOURCE(PCMan_IDC_HAND_CUR))){
			return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
		}else{
			if(wparam & MK_CONTROL){
				return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam ^ MK_CONTROL, lparam);
			}else{
				if(currentSelecting){
					return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
				}else{
					if(wparam & MK_LBUTTON){
						lbtnAndSomeButton = true;
					}else if(wparam & MK_MBUTTON){
						wbtnAndSomeButton = true;
					}else{
						rbtnAndSomeButton = false;
					}
					return 0;
				}
			}
		}
	}else if(msg == WM_RBUTTONUP){
		if(!Settings::getInstance().getEnabledMouseRight() || currentCursor == LoadCursor(NULL, IDC_HAND) || currentCursor == LoadCursor(LoadLibrary(L"PCman.exe"), MAKEINTRESOURCE(PCMan_IDC_HAND_CUR))){
			return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
		}else{
			if(wparam & MK_CONTROL){
				return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam ^ MK_CONTROL, lparam);
			}else{
				if(currentSelecting){
					currentSelecting = false;
					return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
				}
				if(wparam & MK_LBUTTON){
					keystroke(Settings::getInstance().getHKLR().vkeyCode, Settings::getInstance().getHKLR().modifiers);
				}else if(wparam & MK_MBUTTON){
					keystroke(Settings::getInstance().getHKWR().vkeyCode, Settings::getInstance().getHKWR().modifiers);
				}else{
					if(!rbtnAndSomeButton){
						keybd_event((Settings::getInstance().getLRRL() == Settings::LR_TO_RL) ? VK_LEFT : VK_RIGHT, 0, 0, 0);
						keybd_event((Settings::getInstance().getLRRL() == Settings::LR_TO_RL) ? VK_LEFT : VK_RIGHT, 0, KEYEVENTF_KEYUP, 0);
					}
				}
					
				return 0;
			}
		}
	// Mbutton -> end
	}else if(msg == WM_MBUTTONDOWN){
		if(!Settings::getInstance().getEnabledMouseMiddle()){
			return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
		}else{
			if(wparam & MK_CONTROL){
				return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam ^ MK_CONTROL, lparam);
			}else{
				if(wparam & MK_LBUTTON){
					// do nothing
				}else if(wparam & MK_RBUTTON){
					// do nothing
				}else{
					wbtnAndSomeButton = false;
				}
				return 0;
			}
		}
	}else if(msg == WM_MBUTTONUP){
		if(!Settings::getInstance().getEnabledMouseMiddle()){
			return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd,msg,wparam,lparam);
		}else{
			if(wparam & MK_CONTROL){
				return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam ^ MK_CONTROL, lparam);
			}else{
				if(wparam & MK_LBUTTON){
					lbtnAndSomeButton = true;
					keystroke(Settings::getInstance().getHKLW().vkeyCode, Settings::getInstance().getHKLW().modifiers);
				}else if(wparam & MK_RBUTTON){
					rbtnAndSomeButton = true;
					keystroke(Settings::getInstance().getHKRW().vkeyCode, Settings::getInstance().getHKRW().modifiers);
				}else{
					if(!wbtnAndSomeButton){
						keybd_event(VK_END, 0, 0, 0);
						keybd_event(VK_END, 0, KEYEVENTF_KEYUP, 0);
					}
				}
				return 0;
			}
		}
	}else if(msg == WM_MOUSEWHEEL){
		if(!Settings::getInstance().getEnabledMouseWheel()){
			return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
		}else{
			if(LOWORD(wparam) & MK_CONTROL){
				return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam ^ MK_CONTROL, lparam);
			}else{
				if(wparam & MK_LBUTTON){
					lbtnAndSomeButton = true;
				}else if(wparam & MK_RBUTTON){
					rbtnAndSomeButton = true;
				}
				accumulatedDelta += GET_WHEEL_DELTA_WPARAM(wparam);
				if(accumulatedDelta > 0){
					for(int i = 1; i <= accumulatedDelta / WHEEL_DELTA; i++, accumulatedDelta -= WHEEL_DELTA){
						if(wparam & MK_LBUTTON){
							switch(Settings::getInstance().getKeyPairLeftButtonWheelScroll()){
								case Settings::PAGEUPDOWN:
									keybd_event(VK_PRIOR, 0, 0, 0);
									keybd_event(VK_PRIOR, 0, KEYEVENTF_KEYUP, 0);
									break;
								case Settings::BRACKET:
									keybd_event(VK_OEM_4, 0, 0, 0);
									keybd_event(VK_OEM_4, 0, KEYEVENTF_KEYUP, 0);
									break;
								case Settings::FB:
									keybd_event('B', 0, 0, 0);
									keybd_event('B', 0, KEYEVENTF_KEYUP, 0);
									break;
							}
						}else if(wparam & MK_RBUTTON){
							switch(Settings::getInstance().getKeyPairRightButtonWheelScroll()){
								case Settings::PAGEUPDOWN:
									keybd_event(VK_PRIOR, 0, 0, 0);
									keybd_event(VK_PRIOR, 0, KEYEVENTF_KEYUP, 0);
									break;
								case Settings::BRACKET:
									keybd_event(VK_OEM_4, 0, 0, 0);
									keybd_event(VK_OEM_4, 0, KEYEVENTF_KEYUP, 0);
									break;
								case Settings::FB:
									keybd_event('B', 0, 0, 0);
									keybd_event('B', 0, KEYEVENTF_KEYUP, 0);
									break;
							}
						}else{
							keybd_event(VK_UP, 0, 0, 0);
							keybd_event(VK_UP, 0, KEYEVENTF_KEYUP, 0);
						}
					}
					accumulatedDelta %= WHEEL_DELTA;
				}else if(accumulatedDelta < 0){
					for(int i = 1; i <= (-accumulatedDelta) / WHEEL_DELTA; i++, accumulatedDelta += WHEEL_DELTA){
						if(wparam & MK_LBUTTON){
							switch(Settings::getInstance().getKeyPairLeftButtonWheelScroll()){
								case Settings::PAGEUPDOWN:
									keybd_event(VK_SPACE, 0, 0, 0);
									keybd_event(VK_SPACE, 0, KEYEVENTF_KEYUP, 0);
									break;
								case Settings::BRACKET:
									keybd_event(VK_OEM_6, 0, 0, 0);
									keybd_event(VK_OEM_6, 0, KEYEVENTF_KEYUP, 0);
									break;
								case Settings::FB:
									keybd_event('F', 0, 0, 0);
									keybd_event('F', 0, KEYEVENTF_KEYUP, 0);
									break;
							}
						}else if(wparam & MK_RBUTTON){
							switch(Settings::getInstance().getKeyPairRightButtonWheelScroll()){
								case Settings::PAGEUPDOWN:
									keybd_event(VK_SPACE, 0, 0, 0);
									keybd_event(VK_SPACE, 0, KEYEVENTF_KEYUP, 0);
									break;
								case Settings::BRACKET:
									keybd_event(VK_OEM_6, 0, 0, 0);
									keybd_event(VK_OEM_6, 0,KEYEVENTF_KEYUP, 0);
									break;
								case Settings::FB:
									keybd_event('F', 0, 0, 0);
									keybd_event('F', 0, KEYEVENTF_KEYUP, 0);
									break;
							}
						}else{
							keybd_event(VK_DOWN, 0, 0, 0);
							keybd_event(VK_DOWN, 0, KEYEVENTF_KEYUP, 0);
						}
					}
				}
				return 0;
			}
		}			
	}

	return CallWindowProc(reinterpret_cast<WNDPROC>(previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
}

bool CallbackImpl::reappendPCManMenu(HWND hwnd){
	wchar_t className[512] = {0};
	GetClassName(hwnd, className, 511);
	if(wcscmp(className, L"PCMan") == 0 || wcscmp(className, L"PCManCB") == 0 || wcscmp(className, L"PCManC") == 0){
		HMENU mainMenu = GetMenu(hwnd);
		if(!mainMenu){
			ApplicationUtility::getInstance().showLastErrorMessage(L"չϹ PCMan D GetMenu() !", MB_ICONERROR);
		}else{
			// check if the last item is already PPK Mouse Addon
			int i = GetMenuItemCount(mainMenu) - 1;
			MENUITEMINFO mii;
			mii.cbSize = sizeof(MENUITEMINFO);
			wchar_t itemString[512]={0};
			mii.dwTypeData = &itemString[0];
			mii.cch = 512;
			mii.fMask = MIIM_TYPE;
			mii.fType = MFT_STRING;
			if(GetMenuItemInfo(mainMenu, i, TRUE, &mii)){
				if(wcscmp(itemString, L"PPK ƹ~") == 0){
					return true;
				}
			}

			if(!AppendMenu(mainMenu, MF_POPUP, reinterpret_cast<UINT_PTR>(menuAppend = LoadMenu(LoadLibrary(L"msimg32.dll"), MAKEINTRESOURCE(IDR_APPEND_MAINMENU))), L"PPK ƹ~")){
				ApplicationUtility::getInstance().showLastErrorMessage(L"չϹ PCMan D AppendMenu() !", MB_ICONERROR);
			}else{
				if(Settings::getInstance().getEnabled()){
					MENUITEMINFO mii = {0};
					mii.cbSize = sizeof(MENUITEMINFO);
					mii.fMask = MIIM_STATE;
					GetMenuItemInfo(menuAppend, ID_PPK_ENABLED, FALSE, &mii);
					mii.fState |= MFS_CHECKED;
					SetMenuItemInfo(menuAppend, ID_PPK_ENABLED, FALSE, &mii);
				}
				DrawMenuBar(hwnd);
				return true;
			}
		}
	}
	return false;
}

void CallbackImpl::keystroke(WORD vKeyCode, WORD modifiers){
	if(modifiers & HOTKEYF_ALT){
		keybd_event(VK_MENU, 0, 0, 0);
	}
	if(modifiers & HOTKEYF_CONTROL){
		keybd_event(VK_CONTROL, 0, 0, 0);
	}
	if(modifiers & HOTKEYF_SHIFT){
		keybd_event(VK_SHIFT, 0, 0, 0);
	}
	keybd_event(static_cast<BYTE>(vKeyCode), 0, 0, 0);
	keybd_event(static_cast<BYTE>(vKeyCode), 0, KEYEVENTF_KEYUP, 0);
	if(modifiers & HOTKEYF_ALT){
		keybd_event(VK_MENU, 0, KEYEVENTF_KEYUP, 0);
	}
	if(modifiers & HOTKEYF_CONTROL){
		keybd_event(VK_CONTROL, 0, KEYEVENTF_KEYUP, 0);
	}
	if(modifiers & HOTKEYF_SHIFT){
		keybd_event(VK_SHIFT, 0, KEYEVENTF_KEYUP, 0);
	}
}