/*
PPKMouseAddon
 by Mnjul/purincess (Min-Zhong Lu)
 # Middle Sea, Mnjul's Intimate Home

Copyright 2007-2010 Min-Zhong Lu

This file is part of PPKMouseAddon.

PPKMouseAddon is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

PPKMouseAddon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PPKMouseAddon.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __CALLBACK_IMPL_H
#define __CALLBACK_IMPL_H

#include "ApplicationUtility.h"
#include <map>

class CallbackImpl{
private:
	CallbackImpl() : previousMainWndProc(NULL),
	                 previousTerminalViewWndProc(NULL),
	                 hwndMain(NULL),
	                 menuAppend(NULL),
	                 currentCursor(NULL),
	                 currentSelecting(false),
	                 accumulatedDelta(0),
	                 lbtnAndSomeButton(false),
	                 wbtnAndSomeButton(false),
	                 rbtnAndSomeButton(false){
	}
	~CallbackImpl(){
	}
	CallbackImpl(const CallbackImpl&);	//no copy constructor
	CallbackImpl& operator=(const CallbackImpl&);	// no assignment operation constructor

public:
	static CallbackImpl& getInstance(){
		static CallbackImpl instance;
		return instance;
	}

private:
	static void keystroke(WORD vKeyCode, WORD modifiers);
	typedef std::map<HWND, bool> map_hwnd_bool;
	map_hwnd_bool calledFirstTimeShowWindow;

	static const WORD PCMan_ID_CUSTOMIZE = 32834;
	static const WORD PCMan_IDC_MAINTAB = 6;
	static const WORD PCMan_IDC_HAND_CUR = 137;


protected:
	LONG_PTR previousMainWndProc;
	LONG_PTR previousTerminalViewWndProc;

	HWND hwndMain;
	HMENU menuAppend;
	HCURSOR currentCursor;

	POINT mousePosition;
	bool currentSelecting;

	short accumulatedDelta;

	bool lbtnAndSomeButton;
	bool wbtnAndSomeButton;
	bool rbtnAndSomeButton;

	LRESULT callWndProcInternal(int code, WPARAM wparam, LPARAM lparam);
	LRESULT newMainWndProcInternal(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);
	LRESULT newTerminalViewWndProcInternal(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam);

	bool reappendPCManMenu(HWND hwnd);

public:
	static LRESULT CALLBACK callWndProc(int code, WPARAM wparam, LPARAM lparam){
		try{
			return CallbackImpl::getInstance().callWndProcInternal(code, wparam, lparam);
		}catch(unsigned int errCode){
			ApplicationUtility::getInstance().showTypeSafeErrorMessage(errCode, L"CallbackImpl::callWndProc");
			return CallNextHookEx(/* ignored */ NULL, code, wparam, lparam);
		}
	}

	static LRESULT CALLBACK newMainWndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam){
		try{
			return CallbackImpl::getInstance().newMainWndProcInternal(hwnd, msg, wparam, lparam);
		}catch(unsigned int errCode){
			ApplicationUtility::getInstance().showTypeSafeErrorMessage(errCode, L"CallbackImpl::newMainWndProc");
			return CallWindowProc(reinterpret_cast<WNDPROC>(CallbackImpl::getInstance().previousMainWndProc), hwnd, msg, wparam, lparam);
		}
	}

	static LRESULT CALLBACK newTerminalViewWndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam){
		try{
			return CallbackImpl::getInstance().newTerminalViewWndProcInternal(hwnd, msg, wparam, lparam);
		}catch(unsigned int errCode){
			ApplicationUtility::getInstance().showTypeSafeErrorMessage(errCode, L"CallbackImpl::newTerminalViewWndProc");
			return CallWindowProc(reinterpret_cast<WNDPROC>(CallbackImpl::getInstance().previousTerminalViewWndProc), hwnd, msg, wparam, lparam);
		}
	}

};

#endif